// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
/*****************************************************************************
 *
 *  BoolListParameter.cc                        XMM Science Analysis Subsystem
 *
 *  Boolean-std::list task parameter
 *
 *  (c) 1998  ESA  -  Written by Jon Brumfitt, Aurora
 *
 *  $Id: BoolListParameter.cc,v 1.22 2002/12/17 17:53:48 jbakker Exp $
 *
 *****************************************************************************/

#include <sstream>             // C++ headers
#include <iostream>
#include <string>
#include <vector>

#include "BoolListParameter.h"     // Application headers
#include "StringToken.h"
#include <errstr.h>
#include <StringFns.h>


//================================================================
//  Class BoolListParameter
//================================================================

/*
 *  All instances of std::vector<bool> have been replaced by std::vector<int> to
 *  work-around a problem with the egcs compiler. Consequently, setValue()
 *  cannot be used at present (unless a std::vector<int> argument is used).
 */


/*** Format the parameter value as a std::string in quotes ***/

std::string BoolListParameter::asString() const
{
    std::ostringstream ostr;
    std::vector<int>::const_iterator it = _value.begin();
    if(it != _value.end())
        ostr << (*it++?"yes" : "no");
    while(it != _value.end())
        ostr << ' ' << (*it++?"yes" : "no");
    return ostr.str();
}


/*** Format the default value as a std::string in quotes ***/

std::string BoolListParameter::defaultAsString() const
{
    std::ostringstream ostr;
    std::vector<int>::const_iterator it = _default.begin();
    if(it != _default.end())
        ostr << (*it++?"yes" : "no");
    while(it != _default.end())
        ostr << ' ' << (*it++?"yes" : "no");

    return ostr.str();
}


/*
 *  Parse std::string representation and assign to 'value' if it is valid.
 *  The validation is not strictly necessary, since bool must be valid.
 */

bool BoolListParameter::parse(const std::string& str, std::vector<int>& value)
{
    std::list<std::string> stringTokens = StringFns::splitString(std::string(str)," \t");

    bool ok = true;
    std::list<std::string>::const_iterator it = stringTokens.begin();
    while(ok && it != stringTokens.end())
    {
        std::string s = *it++;
	StringToken t(s);
	bool i = t.asBool();
	if(t.isValid())
	    value.push_back(i);
	else {
	  errstr << "Parameter '" << _name << "'" << Sas::error("ParamWrongType");
	  ok = false;
	}
    }
    return ok;
}


/*** Revert to default value ***/

void BoolListParameter::revertToDefault()
{
    _value = _default;
    notifyView();
}


/*** Test whether value is allowable for the parameter ***/

bool BoolListParameter::isValid(const std::vector<int>&) const
{
    return true;
}


/*** Set the parameter to the specified value ***/

bool BoolListParameter::setValue(const std::vector<int>& value)
{
    bool valid = isValid(value);
    if(valid)
    {
        _value = value;
	setDefined();
	notifyView();
    }
    return valid;
}


/*** Set the default to the specified value ***/

bool BoolListParameter::setDefault(const std::vector<int>& value)
{    
    bool valid = isValid(value);
    if(valid)
        _default = value;
    return valid;
}


/*** Set the value, specified as a std::string ***/

bool BoolListParameter::setNewValue(const std::string& str)
{
    std::vector<int> values;
    bool ok =  parse(str, values) && setValue(values);

    if(!ok)
        errstr << "Parameter '" << _name << "' cannot be assigned new value '"
	       << str << "'" << Sas::error("ParamNotAssigned");

    return true;
}


/***  Set the default, specified as a std::string ***/

bool BoolListParameter::setNewDefault(const std::string& str)
{
    std::vector<int> values;
    bool ok = parse(str, values) && setDefault(values);

    if(!ok)
        errstr << "Parameter '" << _name << "' cannot be assigned default value "
	       << str << "'" << Sas::error("ParamNotAssigned");
    return true;
}


/*** Return the length of the std::list ***/

int BoolListParameter::size() const
{
    return _value.size();
}
