// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
/*****************************************************************************
 *
 *  BoolParameter.cc                            XMM Science Analysis Subsystem
 *
 *  Boolean task parameter
 *
 *  (c) 1998  ESA  -  Written by Jon Brumfitt, Aurora
 *
 *  $Id: BoolParameter.cc,v 1.12 2002/12/17 17:53:49 jbakker Exp $
 *
 *****************************************************************************/

#include <iostream>             // C++ headers
#include <string>

#include "BoolParameter.h"        // SAS headers
#include "StringToken.h"
#include <errstr.h>


//================================================================
//  Class BoolParameter
//================================================================

/*** Format the parameter value as a std::string ***/

std::string BoolParameter::asString() const
{
    return (_value ? "yes" : "no");
}


/*** Format the default value as a std::string ***/

std::string BoolParameter::defaultAsString() const
{
    return (_default ? "yes" : "no");
}


/***  Parse std::string representation and assign to 'value' if it is valid ***/

bool BoolParameter::parse(const std::string& str, bool& value)
{
    bool ok = true;
    StringToken t(str);
    bool n = t.asBool();
    if(t.isValid())
        value = n;
    else {
      errstr << "Parameter '" << _name << "'" << Sas::error("ParamWrongType");
      ok = false;
    }
    return ok;
}


/*** Revert to default value ***/

void BoolParameter::revertToDefault()
{
    _value = _default;
    notifyView();
}


/*** Set the parameter to the specified value ***/

bool BoolParameter::setValue(bool value)
{
    _value = value;
    setDefined();
    notifyView();
    return true;
}


/*** Set the default to the specified value ***/

bool BoolParameter::setDefault(bool value)
{
    _default = value;
    return true;
}


/*** Set the value, specified as a std::string ***/

bool BoolParameter::setNewValue(const std::string& str)
{
    bool value = false;
    bool ok = parse(str, value) && setValue(value);

    if(!ok)
      errstr << "Parameter '" << _name << "' cannot be assigned new value '"
	     << str << "'" << Sas::error("ParamNotAssigned");

    return true;
}


/***  Set the default, specified as a std::string ***/

bool BoolParameter::setNewDefault(const std::string& str)
{
    bool value = false;
    bool ok = parse(str, value) && setDefault(value);

    if(!ok)
      errstr << "Parameter '" << _name << "' cannot be assigned default value "
	     << str << Sas::error("ParamNotAssigned");

    return true;
}
