// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
/**-*-C++-*-*******************************************************************
 *
 *  Parameter.h                                 XMM Science Analysis Subsystem
 *
 *  Parameter of a task
 *
 *  (c) 1998  ESA  -  Written mv Deby Jon Brumfitt, Aurora
 *
 *  $Id: Parameter.h,v 1.36 2002/12/17 17:53:49 jbakker Exp $
 *
 *****************************************************************************/

#ifndef PARAMETER_H
#define PARAMETER_H

#include <iostream>             // C++ headers
#include <string>
#include <list>
#include <vector>

#include "ParameterView.h"


//================================================================
//  Abstract Class Parameter
//================================================================

class ChoiceChildParameters;

class Parameter
{
public:
  
  // Type definitions
  enum ParamType{Undefined, 
		 Bool, Int, Real, String, StringChoice, FileName,
		 BoolList, IntList, RealList, StringList, 
		 StringChoiceList, FileNameList,
		 Directory, DataSet, Block, Table, Array, Column, Plane,
		 DirectoryList, DataSetList, BlockList, TableList,
		 ArrayList, ColumnList, PlaneList,
		 Angle, Time, AngleList, TimeList,
		 
		 // added type!
		 DirOrDataSet, DirOrFile
  };
  
  // Constructors & destructors
  Parameter(const std::string& name = "", const std::string& prompt = "");
  virtual ~Parameter();
  
  // Virtual copy constructor
  virtual Parameter* clone() = 0;
  
  // View
  void addView(ParameterView* view);
  void deleteView(ParameterView* view);
  void notifyView();
  
  // Parameter name and prompt message
  const std::string& name()   const { return _name; }
  const std::string& prompt() const { return _prompt; }
  void setName(const std::string& name) { _name = name; }
  void setPrompt(const std::string& prompt) { _prompt = prompt; }
  
  
  // Accessing parameter attributes
  virtual ParamType   getType() const {return Undefined;}
  virtual const char* getTypeName() const { return ""; }
  virtual int         size() const {return -1;}
  
  // Accessing value
  virtual bool           boolValue() const;
  virtual int            intValue() const;
  virtual double         realValue() const;
  virtual const std::string&  stringValue() const;
  virtual std::vector<int>    boolListValue() const;
  virtual std::vector<int>    intListValue() const;
  virtual std::vector<double> realListValue() const;
  virtual std::vector<std::string> stringListValue() const;
  
  // Setting value
  virtual bool setNewValue(const std::string&) {return false;};
  virtual bool setValue(bool) { wrongType(); return false; }
  virtual bool setValue(int) { wrongType(); return false; }
  virtual bool setValue(double) { wrongType(); return false; }
  virtual bool setValue(const std::string&) { wrongType(); return false; }
  virtual bool setValue(const char*) { wrongType(); return false; }
  //  Note: vector<bool> temporarily disabled to work-around egcs problem
  //  virtual bool setValue(const vector<bool>&) { wrongType(); return false; }
  virtual bool setValue(const std::vector<int>&) { wrongType(); return false; }
  virtual bool setValue(const std::vector<double>&) { wrongType(); return false; }
  virtual bool setValue(const std::vector<std::string>&) { wrongType(); return false; }
  bool setValue(const Parameter*);
  
  // Defaults
  virtual bool setNewDefault(const std::string&) {return false;};
  virtual void revertToDefault() {};       // SHOULD BE "= 0"
  
  // Constraints...
  void setConstraints(const std::string&);
  const std::string& constraints() const;
  // ...or choices
  void addChoice(const std::string& c) {addChild(c,(Parameter*)0);}
  std::list<std::string> choices();
  
  // Mandatory parameters
  bool  isSet() const { return _isSet; }
  bool  isMandatory() const { return _isMandatory; }
  void  setMandatory(bool mand) { _isMandatory = mand; }
  
  // Formatting
  virtual std::string asString() const = 0;
  virtual std::string defaultAsString() const { return _emptyString; }
  virtual void writeHelpToStream(std::ostream&) const;
  virtual void writeConstraint(std::ostream&) const;
  friend std::ostream& operator<<(std::ostream& os, const Parameter& p);
  
  // Nesting
  void addChild(const std::string& choice,Parameter* child);
  ChoiceChildParameters&  children() const;
  
protected:
  Parameter(const Parameter&);            // copy constructor is protected!
  Parameter& operator=(const Parameter&); // assignment (disabled)
  
  // Member data
  std::string         _name;
  std::string         _prompt;
  
  // Mandatory parameters
  void setDefined() { _isSet = true; }
  
  virtual bool assignValueOf(const Parameter&)=0;
  
private:
  
  // Error handling
  void wrongType() const;
  
  bool   _isSet;
  bool   _isMandatory;
  std::string _constraints;
  
  // All View available:
  class ParameterViews;
  ParameterViews*  _views;
  ChoiceChildParameters* _children;
  
  
  // Constants
  static const std::string         _emptyString;
  static const std::vector<int>    _emptyBoolList;  // Temporary work-around for egcs
  static const std::vector<int>    _emptyIntList;
  static const std::vector<double> _emptyDoubleList;
  static const std::vector<std::string> _emptyStringList;
};

#endif

