// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
/*****************************************************************************
 *
 *  ParameterInfo.cc                            XMM Science Analysis Subsystem
 *
 *  Information about parameters of a task
 *
 *  (c) 1998  ESA  -  Written by Jon Brumfitt, Aurora
 *
 *  $Id: ParameterInfo.cc,v 1.29 2009/07/02 17:43:38 eojero Exp $
 *
 *****************************************************************************/

#include "ParameterInfo.h"     // Application headers
#include "Parameter.h"
#include "ChoiceChildParameters.h"
#include <errstr.h>

#include <iostream>          // C++ headers
#include <sstream>
#include <cstdlib>


//================================================================
//  Class ParameterInfo
//================================================================

/*** Constructor ***/

ParameterInfo::ParameterInfo()
{
}


/***  Copy constructor (deep copy, including parameters) ***/

ParameterInfo::ParameterInfo(const ParameterInfo& info)
{
    copy(info);
} 


/*** Destructor ***/

ParameterInfo::~ParameterInfo()
{
  for (ParamMap::const_iterator it = _map.begin();it != _map.end();it++)
    delete (*it).second;
}


/*** Assignment operator ***/

ParameterInfo& ParameterInfo::operator=(const ParameterInfo& rhs)
{
    if (this != &rhs) copy(rhs);      
    return *this;
}


/*** Copy [ PRIVATE ] ***/

void ParameterInfo::copy(const ParameterInfo& info)
{
    _taskName = info._taskName;
    _nameList = info._nameList;
    _rootList = info._rootList;

    // copy over basics
    ParameterInfo::ParamMap::const_iterator it;
    for (it=info._map.begin();it!= info._map.end();it++)
      _map[(*it).first]=(*it).second->clone();
    
    // create parent <-> child connections
    for (it=info._map.begin();it!= info._map.end();it++)
    {
      Parameter* newP=_map[(*it).first]; // pointer to Parameter in new std::map;
      Parameter* oldP=(*it).second;      // pointer to Parameter in old std::map;

      // scan for possible children that belong to each choice
      for (ChoiceChildParameters::const_iterator mIt=oldP->children().begin();
	   mIt!=oldP->children().end();mIt++) {
	// 
	const std::string& choice=(*mIt).first;
	newP->addChoice(choice);
	for (std::list<Parameter* >::const_iterator lIt=(*mIt).second.begin();
	     lIt!=(*mIt).second.end();lIt++) {
	  Parameter* child= _map[ (*lIt)->name() ];
	  newP->addChild(choice,child);
	}
      }
    }
    

    
}


/*** Set the name of the task ***/

void ParameterInfo::setTaskName(const std::string& name)
{
    _taskName = name;
}


/*** Write the parameter information to a std::stream ***/

std::ostream& operator<<(std::ostream& os, const ParameterInfo& p)
{
  for (std::list<std::string>::const_iterator lit = p._nameList.begin();
       lit != p._nameList.end(); lit++) {
    
    ParameterInfo::ParamMap::const_iterator mit=p._map.find(*lit);
    if (mit == p._map.end()) errstr << "Could not fetch parameter '" << (*lit)
				  << "' from memory!" 
				  << Sas::error("ParamCorrupted");
    Parameter* p=(*mit).second;
    os << p->name() << " = " << (*p) << std::endl;
  }
  return os;
}


/*** Write the parameter information, including defaults, etc ***/

void ParameterInfo::writeHelpToStream(std::ostream& os) const
{
  // produce list as read from stream.
  for (std::list<std::string>::const_iterator lit = _nameList.begin();
       lit != _nameList.end(); lit++) {

    ParameterInfo::ParamMap::const_iterator mit=_map.find(*lit);
    if (mit == _map.end()) errstr << "Could not fetch parameter '" << (*lit)
				  << "' from memory!" 
				  << Sas::error("ParamCorrupted");
    Parameter* p=(*mit).second;
    p->writeHelpToStream(os);
  }
}


/*** Format the ParameterInfo as a command line ***/

std::string ParameterInfo::asCommandLine() const
{
  std::ostringstream os;
  os << _taskName;

  for (std::list<std::string>::const_iterator lit = _nameList.begin();
       lit != _nameList.end(); lit++) {

    ParameterInfo::ParamMap::const_iterator mit=_map.find(*lit);
    if (mit == _map.end()) errstr << "Could not fetch parameter '" << (*lit)
				  << "' from memory!" 
				  << Sas::error("ParamCorrupted");
    
    Parameter* p=(*mit).second;
    os << " " << p->name() << "=" << (*p);
  }
  //---
  return os.str();
}


/*** Format the ParameterInfo as an argument std::vector ***/

std::vector<std::string> ParameterInfo::asArgumentList() const
{
  std::vector<std::string> arglist;
  arglist.push_back(_taskName);

  for (std::list<std::string>::const_iterator lit = _nameList.begin();
       lit != _nameList.end(); lit++) {
    
    
    ParameterInfo::ParamMap::const_iterator mit=_map.find(*lit);
    if (mit == _map.end()) errstr << "Could not fetch parameter '" << (*lit)
				  << "' from memory!" 
				  << Sas::error("ParamCorrupted");
    Parameter* p=(*mit).second;
    std::string value = p->asString();
    if(value[0] == '\'') value = value.substr(1,value.size()-2);
    arglist.push_back(p->name() + "=" + value);
  }

  return arglist;
}


/*** Parse command line and set values of specified parameters ***/

void ParameterInfo::parseCommandLine(int& argc, char** argv)
{
    char* clmode = getenv("SAS_CLMODE");
    if(clmode)
    {
        std::string s = clmode;
	if(s == "PCS")
	    parsePCSCommandLine(argc, argv);
	else if(s == "SAS")
	    parseSASCommandLine(argc, argv);
	else
	    errstr << "Environment variable SAS_CLMODE = " 
		   << clmode << " not valid" << Sas::error("ParamEnvVariable");
    } else 
        parseSASCommandLine(argc, argv);
}


/*** Parse normal SAS command-line format ***/

void ParameterInfo::parseSASCommandLine(int& argc, char** argv)
{
    for(int i=0; i<argc; i++)
    {
        std::string arg = argv[i];

        // Ignore everything after "--"
	if(arg == "--") break;

        // Only process it when it is a=b style
        unsigned n = arg.find("=");
        if (n == string::npos) continue;

        std::string name;

        // Accept "foo" or "--foo", but ignore "-foo"
	if (arg.substr(0,2) == "--")
            name = arg.substr(2,n-2);
        else if (arg.substr(0,2) == "-")
            continue;
        else
            name = arg.substr(0,n);

        std::string value = arg.substr(n+1,std::string::npos);

        // Ignore parameters which are not in the parameter file
        if(_map.count(name) <= 0) continue;
	
        // Assign value to the parameter
	Parameter *p = _map[name];
        if(p && !p->setNewValue(value))
	    errstr << "Could not assign value " << value << " to " << name << std::endl;
	
        // Indicate the options used by setting them to 0
        argv[i] = 0;
    }

    // Remove all arguments that have been set to 0
    for (int i = 0; i < argc; ) {
        if (argv[i]) 
	    ++i;
        else {
	    for (int j = i; j < argc-1; ++j)
	        argv[j] = argv[j+1];
	    --argc;
	}
    }
}


/*** Parse special PCS command-line format ***/

void ParameterInfo::parsePCSCommandLine(int& argc, char** argv)
{
    typedef std::map<std::string,std::string,std::less<std::string> > OptionMap;
    OptionMap options;
    std::string mode = "odf";           // Default PCS keyword
  
    for (int i = 1; i < argc; ++i)
    {
        std::string arg = argv[i];

	if (arg.substr(0,2) == "--") {
	    // Ignore "--" options

	} else if (arg[0] == '-') {

	    // Switch mode on PCS keywords
	    if (arg == "-coeff" || arg == "-ccf" || arg == "-sp")
	    {
	        mode = arg.substr(1,std::string::npos);
		argv[i] = 0;
	    }
	} else {

	    // Append argument to value for current PCS keyword
	    if (!options[mode].empty()) 
	        options[mode] += " ";
	    options[mode] += argv[i];
	    argv[i] = 0;
	}
    }

    // Add the parameters
    for (OptionMap::const_iterator ix = options.begin(); ix != options.end(); ++ix)
    {
        std::string name  = (*ix).first;
	std::string value = (*ix).second;

        if(_map.count(name) > 0)    // Parameter exists
	{
	    Parameter *p = _map[name];
	    if(p && !p->setNewValue(value))
	        errstr << "Could not assign value " << value << 
		  " to " << name << std::endl;
	}
    }

    // Remove all argument that have been set to 0
    for (int i = 0; i < argc; )
    {
        if (argv[i]) ++i;
	else {
	    for (int j = i; j < argc-1; ++j)
	        argv[j] = argv[j+1];
	    --argc;
	}
    }
}


/*** Add a new parameter ***/

void ParameterInfo::addParameter(Parameter* p,bool root)
{
  const std::string& name=p->name();
  if (_map.find(name) != _map.end())
    errstr << "Cannot create duplicate parameter with name(" << name << ")"
	   << Sas::error("ParamDuplicate");
  _map[name]=p;
  
  std::list<std::string>::const_iterator it = _nameList.begin();
  while (it != _nameList.end() && (*it) != name) it++;
  if (it == _nameList.end())  {
    _nameList.push_back(name);
    if (root) _rootList.push_back(name);
  }
}


/*** Return the std::list of parameter names (in order) ***/

std::list<std::string> ParameterInfo::names() const
{
    return _nameList;
}
std::list<std::string> ParameterInfo::rootNames() const
{
    return _rootList;
}
 

/*** Return the parameter "name" ***/

Parameter* ParameterInfo::operator[](const std::string& name) const
{
    return parameter(name, true);
}


/*** Return the parameter "name" ***/

Parameter* ParameterInfo::parameter(const std::string& name, bool alertError) const
{
    if(_map.count(name) > 0)
        return _map[name];
    else {
        if(alertError)
  	    errstr << "Parameter '" << name << "' not found in parameter space of '"
		   << _taskName << "'"
		   << Sas::error("ParamNotFound");
        return 0;
    }
}


/*** Revert each parameter to its default value ***/

void ParameterInfo::revertToDefault()
{
    std::list<std::string>::iterator iter = _nameList.begin();
    while(iter != _nameList.end())
    {
        Parameter* p = parameter(*iter++,true);
        if(p)
            p->revertToDefault();
    }
}


/** Check that all mandatory parameters have been set ***/

static void checkMandatoryChilds(std::string& missing,Parameter* p, Parameter* parent=0)
{
  if (!p) return;
  if(p->isMandatory() && !p->isSet()) {
    if (!missing.empty()) missing += ", ";
    missing += p->name();
    if (parent) missing += " (triggered by " + parent->name() + ')';
  }

  else {
    // hack...
    std::string value = p->asString();
    if(value[0] == '\'') value = value.substr(1,value.size()-2);

    // ... may have children
    ChoiceChildParameters::const_iterator cIt=
      p->children().find(value);
    if (cIt == p->children().end()) return;

    for (std::list<Parameter* >::const_iterator it=(*cIt).second.begin();
	 it!=(*cIt).second.end();it++) {
      checkMandatoryChilds(missing,*it,p);
    }
  }
}

void ParameterInfo::checkMandatoryParameters()
{
    std::string missing;
    for (std::list<std::string>::const_iterator it=_rootList.begin();
	 it!=_rootList.end();it++) 
      checkMandatoryChilds(missing,parameter(*it));

    if (!missing.empty())
    {
        errstr << "Please specify mandatory parameters:'"
	       << missing 
	       << "' belonging to parameter space of '"
	       << _taskName << "'"
	       << Sas::error("ParamMandatory");
    }
}
