// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
/*****************************************************************************
 *
 *  StringListParameter.cc                      XMM Science Analysis Subsystem
 *
 *  List task parameter
 *
 *  (c) 1998  ESA  -  Written by Jon Brumfitt, Aurora
 *
 *  $Id: StringListParameter.cc,v 1.21 2009/07/02 17:43:38 eojero Exp $
 *
 *****************************************************************************/

#include <sstream>                // C++ headers
#include <iostream>
#include <string>
#include <vector>
#include <list>
#include <cstring>

#include "StringListParameter.h"      // Application headers
#include "QuotedString.h"
#include "ChoiceChildParameters.h"
#include "StringToken.h"
#include <errstr.h>
#include <StringFns.h>
#include <Evaluator.h> // selectlib


std::string StringListParameter::asString() const
{
    std::ostringstream ostr;
    std::vector<std::string>::const_iterator it = _value.begin();
    if(it != _value.end())
      ostr << QuotedString(*it++);
    while(it != _value.end())
      ostr << ' ' << QuotedString(*it++);
    return ostr.str();
}


std::string StringListParameter::defaultAsString() const
{
    std::ostringstream ostr;
    std::vector<std::string>::const_iterator it = _default.begin();
    if(it != _default.end())
      ostr << QuotedString(*it++);
    while(it != _default.end())
      ostr << ' ' << QuotedString(*it++);
    return ostr.str();
}


bool StringListParameter::parse(const std::string& str, std::vector<std::string>& value)
{
    std::list<std::string> tokens = StringFns::splitStringList(std::string(str));

    bool ok = true;
    std::list<std::string>::const_iterator it = tokens.begin();
    while(ok && it != tokens.end())
    {
        std::string s = *it++;
	StringToken t(s);
	std::string i = t.asString();
	if(t.isValid())
	    value.push_back(i);
	else {
	  errstr << "Parameter '" << _name << "'" << Sas::error("ParamWrongType");
	  ok = false;
	}
    }
    return ok;
}



void StringListParameter::revertToDefault()
{
    _value = _default;
    notifyView();
}


bool StringListParameter::isValid(const std::vector<std::string>& v) const
{

  // constraints:
  Evaluator evaluator( constraints() );
  StringSymbols ids = evaluator.stringIdentifiers();  
      
  if (ids.textPtr.size() == 0) {
    if (!evaluator.evalLogic()) return false;
  }
  else {
    
    for (std::vector<std::string>::const_iterator it = v.begin();it != v.end();it++) {
      const char* value=(*it).c_str();
      unsigned int length=(*it).length();
      for (unsigned int i=0;i<ids.textPtr.size();i++) {
	
	/* temporary hack to ensure the fault in selectlib is not corrupting
	   memory! TODO: fix, once selectlib has reimplemeted this.
	*/
	if (ids.maxLength[i] < length) 
	  errstr << "Cannot evaluate \""<< name() << "=" << value << "\", as value is too large for selectlib" << Sas::error("ParamEvalPanic");
	
	strncpy(ids.textPtr[i], value, ids.maxLength[i]);
      }
      if (!evaluator.evalLogic()) return false;
    }
  }

  // children:
  if (!children().empty()) {
    for (std::vector<std::string>::const_iterator it = v.begin();
	 it != v.end();it++) {
      
      if ( children().find(*it) == children().end() ) return false;
    }
  }

  return true;
}


bool StringListParameter::setValue(const std::vector<std::string>& value)
{
    bool valid = isValid(value);
    if(valid)
    {
        _value = value;
	setDefined();
	notifyView();
    }    
    return valid;
}


bool StringListParameter::setDefault(const std::vector<std::string>& value)
{
    bool valid = isValid(value);
    if(valid)
        _default = value;
    return valid;
}


bool StringListParameter::setNewValue(const std::string& str)
{
    std::vector<std::string> values;
    bool ok =  parse(str, values) && setValue(values);

    if(!ok)
        reportRangeError(str);

    return true;
}


bool StringListParameter::setNewDefault(const std::string& str)
{    
    std::vector<std::string> values;
    bool ok = parse(str, values) && setDefault(values);

    if(!ok)
        reportRangeError(str);

    return true;
}


int StringListParameter::size() const
{
    return _value.size();
}


std::list<std::string> StringListParameter::splitStringList(const std::string& str)
{
    std::list<std::string> strlist;
    const char* p = str.c_str();
    char c =*p++;

    while(c)
    {
        std::string token = "";

        // Skip leading space
	while(c && (c ==' ' || c == '\t')) 
	    c=*p++;

	// Quotes
	if(c == '"' || c == '\'') {
	    char quote = c;
	    c=*p++;
	    while(c && (c != quote)) {
	        token += c;
		c=*p++;
	    }
	    if(c != quote)
	      errstr << "Mismatched quotes for parameter '"
		     << _name << "' at token["<< strlist.size()+1 << "]="
		     << token
		     << Sas::error("ParamWrongQuotes");
	    c=*p++;
	    strlist.push_back(token);
	} 
	else if(c && (c != ' ' && c != '\t')) 
	{
	    // Simple token
	    while(c && (c != ' ' && c != '\t')) {
	        token += c;
		c=*p++;
	    }
	    strlist.push_back(token);
	}
    } 
    return strlist;
}


void StringListParameter::reportRangeError(const std::string& valueString)
{
  errstr << "Parameter '" << _name << "';" << std::endl
	 << "value '" << valueString << "' is outside range ";

  if (children().empty()) {
    writeConstraint(errstr);
  }
  else {
    errstr << '\'';
    bool next=false;
    for (ChoiceChildParameters::const_iterator it=children().begin();
       it !=children().end();it++) {
      if (next) errstr << '|';
      next=true;
      errstr << (*it).first;
    }
    errstr << '\'';    
  }
  errstr << Sas::error("ParamNotInRange");
}
