// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
/*****************************************************************************
 *
 *  StringParameter.cc                          XMM Science Analysis Subsystem
 *
 *  String task parameter
 *
 *  (c) 1998  ESA  -  Written by Jon Brumfitt, Aurora
 *
 *  $Id: StringParameter.cc,v 1.15 2002/12/17 17:53:50 jbakker Exp $
 *
 *****************************************************************************/

#include <string>                      // C++ headers
#include <cstring>		       // strncpy

#include "StringParameter.h"           // Application headers
#include "ChoiceChildParameters.h"
#include <errstr.h>
#include <Evaluator.h> // selectlib


std::string StringParameter::asString() const
{
  return _value;
}


std::string StringParameter::defaultAsString() const
{
  return _default;
}


bool StringParameter::parse(const std::string& str, std::string& value)
{
    value = str;
    return true;
}


void StringParameter::revertToDefault()
{
    _value = _default;
    notifyView();
}

bool StringParameter::isValid(const std::string& v) const
{
  const char* value=v.c_str();
  unsigned int length=v.length();

  // constraints:
  Evaluator evaluator( constraints() );
  StringSymbols ids = evaluator.stringIdentifiers();  
  for (unsigned int i=0;i<ids.textPtr.size();i++) {

    /* temporary hack to ensure the fault in selectlib is not corrupting
       memory! TODO: fix, once selectlib has reimplemeted this.
     */
    if (ids.maxLength[i] < length) 
      errstr << "Cannot evaluate \""<< name() << "=" << value << "\", as value is too large for selectlib" << Sas::error("ParamEvalPanic");

    strncpy(ids.textPtr[i], value, ids.maxLength[i]);
  }

  if (!evaluator.evalLogic()) return false;


  // children:
  if (children().empty()) return true;
  for (ChoiceChildParameters::const_iterator it=children().begin();
       it !=children().end();it++) {
    if( (*it).first == v) return true;
  }
  return false;
}


bool StringParameter::setValue(const std::string& value)
{
    bool valid = isValid(value);
    if(valid)
    {
        _value = value;
	setDefined();
	notifyView();
    }
    return valid;
}


bool StringParameter::setDefault(const std::string& value)
{
    bool valid = isValid(value);
    if(valid)
        _default = value;

    return valid;
}


bool StringParameter::setNewValue(const std::string& str)
{
    std::string value;
    bool ok = parse(str, value) && setValue(value);

    if(!ok)
        reportRangeError(str);

    return true;
}


bool StringParameter::setNewDefault(const std::string& str)
{    
    std::string value;
    bool ok = parse(str, value) && setDefault(value);

    if(!ok)
        reportRangeError(str);
    return true;
}


void StringParameter::reportRangeError(const std::string& valueString)
{
  errstr << "Parameter '" << _name << "';" << std::endl
	 << "value '" << valueString << "' is outside range ";
  if (children().empty()) {
    writeConstraint(errstr);
  }
  else {
    errstr << '\'';
    bool next=false;
    for (ChoiceChildParameters::const_iterator it=children().begin();
       it !=children().end();it++) {
      if (next) errstr << '|';
      next=true;
      errstr << (*it).first;
    }
    errstr << '\'';    
  }
  errstr << Sas::error("ParamNotInRange");
}
