// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
/*****************************************************************************
 *
 *  TimeFormatter                               XMM Science Analysis Subsystem
 *
 *  Time with methods for parsing and formatting
 *
 *  (c) 1999  ESA  -  Written by Jon Brumfitt, Aurora
 *
 *  $Id: TimeFormatter.h,v 1.6 2002/12/17 17:53:50 jbakker Exp $
 *
 *****************************************************************************/

#ifndef TIMEFORMATTER_H
#define TIMEFORMATTER_H

#include <string>              // C++ headers
#include <sstream>


//================================================================
//  Class TimeFormatter
//
//  TimeFormatter handles parsing and unparsing of time/date values
//  in a variet of formats:
//
//     xxxx-xx-xxTxx:xx[:xx.xxx]  (FITS date and time)
//     jdxxxxxx.xxx               (Julian day)
//     mjdxx.xxx                  (modified Julian day)
//     yxxxx.xxx                  (decimal year - Not yet implemented)
//     [-]xxxx[.xxx]              (seconds since XMM reference time)
//     xxx xxx xx xx:xx:xx xxxx   (calendar format)
//
//  where `x' is a single digit, `m','j',`d' and `T' are literal characters
//  and square brackets denote optional items. The number of digits shown
//  is not significant.
//
// Examples:
//
//     1997-10-13T00:00:00.000    (FITS date and time)
//     jd2450734.5                (Julian day)
//     mjd50734                   (Modified Julian day)
//     -6.912e+06                 (seconds since XMM Mission Reference Time)
//     Mon Oct 13 00:00:00 1997   (calendar format)
//
//================================================================

class TimeFormatter
{
public:
  enum TimeMode { MRT,  // Mission Reference Time    "123.45"
		  JD,   // Julian Day                "jd450734.5"
		  MJD,  // Modified Julian Day       "mjd50734"
		  CAL,  // Calendar Day              "Sun Sep 16 01:03:52 1973"
		  FITS, // FITS Time                 "1997-10-13T00:00:00.000"
		  DYN   // Decimal Year Number       "y1999.45"
  };

    // Constructor
    TimeFormatter();
    TimeFormatter(TimeMode mode);

    // Format the time as a string
    std::string asString(double value) const;

    // Parse the string and set the time
    bool parse(const std::string& str, double& value);

    void setMode(TimeMode mode=MRT);
    TimeMode mode() { return _mode; }

private:

    // Member data
    TimeMode _mode;


    // Constants

#if __GNUC__ >= 6
    static constexpr double JDMIN = 2400000.0;
#else
    static const double JDMIN = 2400000.0;
#endif

};

#endif

