// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
// -*-C++-*-
//
// File:      XmlParamParser.cc
// Author:    Jorgo Bakker (jbakker@rssd.esa.int)
// Generated: Mon Feb 10 12:48:04 CET 2003
// Usage:     -
// Info:      -

#include "XmlParamParser.h"
#include "ParameterInfo.h"
#include "Parameter.h"      
#include "BoolParameter.h"      
#include "IntParameter.h"      
#include "RealParameter.h" 
#include "AngleParameter.h"
#include "TimeParameter.h"     
#include "StringParameter.h"
#include "SpecialStringParameters.h"
#include "StringParameter.h"
#include "BoolListParameter.h"
#include "IntListParameter.h"
#include "RealListParameter.h"
#include "AngleListParameter.h"
#include "TimeListParameter.h"
#include "StringListParameter.h"
#include "SpecialStringListParameters.h"
#include "StringToken.h"
#include <errstr.h>
#include <ExceptionErrorHandler.h>
#include <StringFns.h>
#include <Evaluator.h> // selectlib
#include <fstream>

#include <qdom.h>
#include <qfile.h>


/* --- syntax of file
 * <FILE>
 *  <CONFIG>
 *    <PARAM .../>
 *    <PARAM .../>
 *  </CONFIG>
 * </FILE>
 */

/* --- syntax of PARAM
 * <PARAM id=ID type=TYPE [list=LIST] [mandatory=MANDATORY] [default=DEFAULT]>
 *   <DESCRIPTION>TEXT</DESCRIPTION>
 *   <CONSTRAINTS>EXPRESSION</CONSTRAINTS>
 *   <CASE>
 *   <ITEM value=VALUE>
 *      <PARAM/>
 *      <PARAM/>
 *   <ITEM value=VALUE>
 *   </CASE>
 * </PARAM>
 */

/* where:
 *  ID= required: name of parameter
 *  TYPE= required: bool|std::string|... 
 *  LIST= optional: yes|no
 *  MANDATORY= optional: yes|no
 *  DEFAULT=optional: default value
 *  TEXT=optional: description of this parameter
 *  VALUE=required: possible value of this parameter
 *        if parameter has VALUE, optional parameters become active
 *  EXPRESSION= select lib expression
 */


void sasQtMsgHandler(QtMsgType, const char* msg)
{
  errstr << msg << Sas::error("QtError");
  /*
  switch (t) {
  case QtWarningMsg:
    errstr << msg << Sas::warning("XmlWarning");
    break;
  case QtFatalMsg:
    errstr << msg << Sas::error("XmlFatal");
    break;
  default:
     errstr << msg << Sas::message(Msg::Core,Msg::Noisy);
    break;
  }
  */
}
//typedef void (*msg_handler)(QtMsgType, const char *);
//Q_EXPORT msg_handler qInstallMsgHandler( msg_handler );


class XmlParser
{
public:
  XmlParser(const std::string& message="")
    : _info(0), _message(message) {}

  ParameterInfo* parse(const QDomDocument& doc);
  
private:
  void handleParameters(const QDomElement& e, Parameter* parent=0, const std::string& item ="");
  void handleCase(const QDomElement& e, Parameter* parent);
  Parameter* constructParameter(const QDomElement& e);
  
  ParameterInfo* _info;
  std::string         _message;
};

ParameterInfo* XmlParser::parse(const QDomDocument& d)
{
  if (d.documentElement().tagName() != "FILE")
	errstr << "Invalid root element <" << d.documentElement().tagName()
	       << "> encountered in parameter file: " 
	       << _message << Sas::error("ParamInvalidElement");
    

  _info = new ParameterInfo;
  
  // find 'tag' CONFIG
  QDomNode n = d.documentElement().firstChild();
  while( !n.isNull() ) {
    if (n.isElement()) {
      QDomElement e=n.toElement();
      if (e.tagName() =="CONFIG") {
	handleParameters(e);
	break;
      }
    }
    n = n.nextSibling();
  }
  return _info;
}


/*
  An element only contains (a list of) <PARAM></PARAM>
*/
void XmlParser::handleParameters(const QDomElement& root,
				 Parameter* parent,
				 const std::string& item
				 )
{
  // check this element for child parameters
  QDomNode n = root.firstChild(); // root=CONFIG|ITEM
  while( !n.isNull() ) {
    if (n.isComment()) {
      ; // ignore
    }
    else if (n.isElement()) {
      QDomElement child=n.toElement();
      
      if (child.tagName() == "PARAM") {
	// construct parameter...
	Parameter* p=constructParameter(child);

	// ...add it to the parameter space
	_info->addParameter(p, parent==0);

	if (parent) parent->addChild(item,p);	
      }
      else 
	errstr << "Invalid construct <" << child.tagName()
	       << "> encountered in parameter file: " 
	       << _message << Sas::error("ParamInvalidElement");
      
    }
    else {
      errstr << "Invalid syntax <" << n.nodeName()
	     << "> encountered in parameter file: " 
	     << _message << Sas::error("ParamInvalidSyntax");
    }
    n = n.nextSibling();
  }
}


/*
  An element only contains (a list of) <PARAM></PARAM>
*/
void XmlParser::handleCase(const QDomElement& root, Parameter* parent)
{
  // check this elements for child parameters
  QDomNode n = root.firstChild(); // <CASE>
  while( !n.isNull() ) {
    if (n.isComment()) {
      ; // ignore
    }
    else if (n.isElement()) {
      QDomElement child=n.toElement(); // <ITEM value='...'>
      if  (child.tagName() != "ITEM")
	errstr << "Invalid child <"<< child.tagName() << "> for "
	       << "<CASE> construct in parameter '" 
	       << parent->name() << "'"
	       << Sas::error("ParamInvalidProperty");

      
      if (!child.hasAttribute("value"))
	errstr << "Missing attribute 'value' of <ITEM> for "
	       << "<CASE> construct in parameter '" 
	       << parent->name() << "'"
	       << Sas::error("ParamInvalidItem");

      // catch invalid entries, and add this choice
      std::string item=child.attribute("value").latin1();
      try {
	parent->addChoice(item);
	parent->setNewDefault(item);
      }
      catch (SasException::Exception) {
	errstr << "Invalid value='"<<item<<"' of <ITEM> for "
	       << "<CASE> construct in parameter '" 
	       << parent->name() << "'"
	       << Sas::error("ParamInvalidItem");
      }

      // check-out value-dependant children
      handleParameters(child,parent,item);

    }
    else {
      errstr << "Invalid syntax <" << n.nodeName()
	     << "> encountered in parameter file: " 
	     << _message << Sas::error("ParamInvalidSyntax");
    }
    n = n.nextSibling();
  }
}


Parameter* XmlParser::constructParameter(const QDomElement& e)
{

  // parameter id, required!
  QString name=e.attribute("id");
  if (name.isNull())
    errstr << "Missing required attribute 'id'"
	   << Sas::error("ParamMissingAttribute");

  // check all valid attributes:
  {
    QDomNamedNodeMap m=e.attributes();
    for (uint i=0;i<m.length();i++) {
      QDomAttr a=m.item(i).toAttr();
      if (!(a.name() == "type" || 
	    a.name() == "id" ||
	    a.name() == "mandatory" ||
	    a.name() == "default" ||
	    a.name() == "list")) 
	
	errstr << "Attribute '" << a.name() << "' of parameter '" << name
	       << "' is not allowed."
	       << Sas::error("ParamInvalidAttribute");
    }
  }

  
  // type, required!
  QString type=e.attribute("type");
  if (type.isNull())
    errstr << "Missing required attribute 'type'"
	   << Sas::error("ParamMissingAttribute");
  bool mayHaveChoice=(type == "string" || type == "bool") ;
  
  // default setting, optional
  bool hasDefault=e.hasAttribute("default");
  QString def=e.attribute("default");

  // mandatory, optional: default=no
  bool isMandatory=false;
  if (e.hasAttribute("mandatory")) {
    QString s=e.attribute("mandatory");
    StringToken t(s.latin1());
    isMandatory=t.asBool();
    if (!t.isValid())
      errstr << "Attribute 'mandatory' of parameter '" << name
	     << "' has invalid boolean value."
	     << Sas::error("ParamInvalidAttribute");
  }
  
  // is a list, optional: default=no
  bool isList=false;
  if (e.hasAttribute("list")) {
    QString s=e.attribute("list");
    StringToken t(s.latin1());
    isList=t.asBool();
    if (!t.isValid())
      errstr << "Attribute 'list' of parameter '" << name
	     << "' has invalid boolean value."
	     << Sas::error("ParamInvalidAttribute");
  }
  
  
  /* valid properties:
     - constraints
     - description
     - case
  */
  std::string description;
  std::string constraints("true"); // default value if not given!
  QDomElement caseElement;
  QDomNode n = e.firstChild();
  while( !n.isNull() ) {
    if (n.isElement()) {
      QDomElement c=n.toElement();
      QString tag=c.tagName();
      if (tag == "DESCRIPTION") description=c.text().stripWhiteSpace().latin1();
      else if (tag == "CONSTRAINTS") {
	constraints=c.text().stripWhiteSpace().latin1();
	if (constraints.empty()) 
	  errstr << "Parameter '" << name << "' contains a <CONSTRAINTS> "
		 << "property, with no data."
		 << Sas::error("ParamInvalidConstraint");
	  
	if (!Evaluator::isValidExpression(constraints)) {
	  errstr << "Parameter '" << name << "' contains a <CONSTRAINTS> "
		 << "property, with invalid selectlib expression: "
		 << constraints
		 << Sas::error("ParamInvalidConstraint");
	}
      }
      else if (tag == "CASE") {
	if (mayHaveChoice) caseElement=c;
	else
	  errstr << "Parameter '" << name << "' is not allowed to contain "
		 << "a <CASE> property."
		 << Sas::error("ParamInvalidProperty");
      }
      else // not a valid property
	errstr << "Parameter '" << name << "' contains an invalid <"
	       << tag << "> property."
	       << Sas::error("ParamInvalidProperty");
    }
    else if (!n.isComment())
      errstr << "Parameter '" << name << "' contains an invalid <"
	     << n.nodeName() << "> node."
	     << Sas::error("ParamInvalidEntity");

    n = n.nextSibling();
  }
  
    
  // create parameter
  Parameter* p=0;
  if      (type == "bool")
    if(isList) p=new BoolListParameter(); else p= new BoolParameter();
  else if (type == "int")
    if(isList) p=new IntListParameter(); else p= new IntParameter();
  else if (type == "real")
    if(isList) p=new RealListParameter(); else p= new RealParameter();
  else if (type == "angle")
    if(isList) p=new AngleListParameter(); else p= new AngleParameter();
  else if (type == "time")
    if(isList) p=new TimeListParameter(); else p= new TimeParameter();
  else if (type == "string")
    if(isList) p=new StringListParameter(); else p= new StringParameter();
  else if (type == "directory")
    if(isList) p=new DirectoryListParameter(); else p= new DirectoryParameter();
  else if (type == "dataset")
    if(isList) p=new DataSetListParameter(); else p=new DataSetParameter();
  else if (type == "directory|dataset" && !isList)
    p = new DirOrDataSetParameter();
  else if (type == "directory|file" && !isList)
    p = new DirOrFileParameter();
  else if (type == "file")
    if(isList) p=new FileNameListParameter(); else p= new FileNameParameter();
  else if (type == "block")
    if(isList) p=new BlockListParameter(); else p= new BlockParameter();
  else if (type == "table")
    if(isList) p=new TableListParameter(); else p= new TableParameter();
  else if (type == "array")
    if(isList) p=new ArrayListParameter(); else p= new ArrayParameter();
  else if (type == "column")
    if(isList) p=new ColumnListParameter(); else p= new ColumnParameter();
  else if (type == "plane")
    if(isList) p=new PlaneListParameter(); else p= new PlaneParameter();
  else {
    errstr << "Attribute 'type' of parameter '" << name
	   << "' has invalid value."
	   << Sas::error("ParamInvalidAttribute");
  }
  errstr << "... constructed parameter(" << name << ','<<type<<')'
	 << Sas::message(Msg::Lib, Msg::Noisy);
  
  // now feed the information into the parameter...
  p->setName(name.latin1());
  p->setMandatory(isMandatory);
  p->setPrompt(description);
  p->setConstraints(constraints);
  if (!caseElement.isNull()) handleCase(caseElement,p);
  if (hasDefault) p->setNewDefault(def.latin1());
  errstr << "... populated parameter(" << name << ','<<type<<')'
	 << Sas::message(Msg::Lib, Msg::Noisy);

  p->revertToDefault();
  errstr << "... initialised parameter(" << name << ','<<type<<')'
	 << Sas::message(Msg::Lib, Msg::Noisy);

  return p;
}


XmlParamParser::XmlParamParser() 
{
}

XmlParamParser::~XmlParamParser()
{
}

ParameterInfo* XmlParamParser::load(const std::string& nameOfFile)
{

  errstr << "Using parameter file " <<  nameOfFile
	 << Sas::message(Msg::Lib, Msg::Noisy);
  ParameterInfo* result=0;
  
  QFile file(nameOfFile.c_str());  
  if (file.open( IO_ReadOnly)) {
    
    /*
      QDomDocument may produce parse errors; we want to catch these errors
      and convert them to Sas errors!
    */
    
    // setup error-bypass
    ExceptionErrorHandler local;
    local.name(errHandler.name());
    ErrorHandler* oldSasHandler = errHandler.handler(&local);
    msg_handler oldQtHandler=qInstallMsgHandler(sasQtMsgHandler);
    
    // now try to parse the file 
    QDomDocument doc(nameOfFile.c_str());
    try {
      if (!doc.setContent(&file)) {
	file.close();
	errstr << "Could not parse contents of xml-file '"
	       << Sas::error("ParamFileNotXml");
      }
      file.close();
      errstr << "... file has proper XML syntax"
	     << Sas::message(Msg::Lib, Msg::Noisy);

      
      // now parse the valid XML file to figure out the SAS parameters:
      XmlParser p(nameOfFile);
      result=p.parse(doc);
      if (result) errstr << "... is proper SAS parameter file"
			<< Sas::message(Msg::Lib, Msg::Noisy);
    }
    catch (SasException::Exception e) {
      qInstallMsgHandler(oldQtHandler);
      errHandler.handler(oldSasHandler);
      errstr << nameOfFile << " : " << e.message
	     << Sas::error("ParamFileInvalid");
    }
    catch (...) {
      qInstallMsgHandler(oldQtHandler);
      errHandler.handler(oldSasHandler);
      errstr << nameOfFile << " : Unexpected exception thrown"
	     << Sas::fatal("UnknownException");
    }
    
    // restore error-bypass
    qInstallMsgHandler(oldQtHandler);
    errHandler.handler(oldSasHandler);
    errstr << "... is proper XML file"
	   << Sas::message(Msg::Lib, Msg::Noisy);
  } else {
    errstr << "Could not open parameter file '"
	   << nameOfFile  << "'"
	   << Sas::error("ParamFileOpen");
  }
  
  return result;
}


/*---                 BEGIN 'helpers for saving xml'                      ---*/

/**
   transforms a string to escaped HTML!
*/
static std::string toXml(const std::string& input)
{
  const char* items="\"&<>";
  std::string result(input);
  std::string::size_type idx=result.find_first_of(items);
  while(idx != std::string::npos) {
    std::string replace;
    switch(result[idx]) {
    case '"': replace="&#34;"; break; // quotation mark = U+0022 ISOnum => quot
    case '&': replace="&#38;"; break; // ampersand, U+0026 ISOnum => amp
    case '<': replace="&#60;"; break; // std::less-than sign, U+003C ISOnum => lt
    case '>': replace="&#62;"; break; // greater-than sign, U+003E ISOnum => gt    
    default: break;
    };
    if (!replace.empty()) result.replace(idx,1,replace);
    idx=result.find_first_of(items,idx+replace.length());
  }
  return result;
}

static void toXml(std::ostream& os,Parameter* p)
{
  // retrieve choices...
  std::list<std::string> choices;
  if (p->getType() == Parameter::Bool) {
    choices.push_back("yes");
    choices.push_back("no");
  }
  choices=p->choices();

  // create header
  os << "  <PARAM"
     << " id=\"" << p->name() << "\""
     << " type=\"" << p->getTypeName() << "\"";
  if (p->size() >= 0)
    os << " list=\"yes\"";
  if (p->isMandatory())
    os << " mandatory=\"yes\"";

  if (p->isSet()) 
    os << " default=\"" <<  toXml(p->asString()) << "\"";
  else if (!p->defaultAsString().empty())
    os << " default=\"" <<  toXml(p->defaultAsString()) << "\"";
  os << ">" << std::endl;

  // create contents
  os << "   <DESCRIPTION>\n    " << toXml(p->prompt()) << "\n"
     << "   </DESCRIPTION>" << std::endl;

  if (choices.empty()) {
    os << "   <CONSTRAINTS>\n    " <<  toXml(p->constraints()) << "\n"
       << "   </CONSTRAINTS>" << std::endl;
  } else {
    os << "   <CASE>\n";
    for (std::list<std::string>::iterator it=choices.begin();it!=choices.end();it++) {
      os << "    <ITEM value=\"" <<  toXml(*it) << "\">\n"
	 << "    </ITEM>\n";
    }
    os << "   </CASE>\n";
  }

  // create footer
  os << "  </PARAM>\n" << std::endl;
}

/*---                 BEGIN 'helpers for saving xml'                      ---*/


void XmlParamParser::save(const ParameterInfo& info,const std::string& nameOfFile)
{
  // create output stream
  std::ofstream os(nameOfFile.c_str());

  if (!os) errstr << "Could not open file \'" << nameOfFile
		  << "\" for writing. Does the directory exist?"
		  << Sas::error("PermissionDenied");

  // create header
  os << "<!-- saved settings for task \""
     << info.taskName() << "\"-->\n" 
     << "<FILE>\n"
     << " <CONFIG>\n"
     << std::endl;

  // save all parameters
  std::list<std::string> parameters=info.names();
  for (std::list<std::string>::iterator it=parameters.begin();
       it != parameters.end();it++){
    Parameter* p=info[*it];
    if (!p) errstr << "Parameter[" << (*it) << "] not accessible"
		   << Sas::fatal("MemoryCorrupted");
    toXml(os,p);
  }
  
  // create footer
  os << " </CONFIG>\n"
     << "</FILE>"
     << std::endl;  
}
